#include <windows.h>
#include <commdlg.h>
#include <stdlib.h>
#include <stdio.h>

#include "Polygon.h"
#include "DXobject.h"
#include "File_io.h"

static OPENFILENAME ofn;

void PopFileInitialize(HWND hwnd)
{
	static char szFilter[] = "Data Explorer (*.dx)\0*.dx\0" \
							 "3D Object Files (*.3do)\0*.3do\0" \
							 "All Files (*.*)\0*.*\0\0";

	ofn.lStructSize			= sizeof(OPENFILENAME);
	ofn.hwndOwner			= hwnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= szFilter;
	ofn.lpstrCustomFilter	= NULL;
	ofn.nMaxCustFilter		= 0;
	ofn.nFilterIndex		= 0;
	ofn.lpstrFile			= NULL;  //set in open and close func
	ofn.nMaxFile			= _MAX_PATH;
	ofn.lpstrFileTitle		= NULL;	//set in open and close func
	ofn.nMaxFileTitle		= _MAX_FNAME + _MAX_EXT;
	ofn.lpstrInitialDir		= NULL;
	ofn.lpstrTitle			= NULL;
	ofn.Flags				= 0;	//set in open and close func
	ofn.nFileOffset			= 0;
	ofn.nFileExtension		= 0;
	ofn.lpstrDefExt			= "dx";
	ofn.lCustData			= 0L;
	ofn.lpfnHook			= NULL;
	ofn.lpTemplateName		= NULL;
}

BOOL PopFileOpenDlg(HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName)
{
	static char szFilter[] = "Data Explorer (*.dx)\0*.dx\0" \
							 "3D Object Files (*.3do)\0*.3do\0" \
							 "All Files (*.*)\0*.*\0\0";

	ofn.hwndOwner			= hwnd;
	ofn.lpstrFilter			= szFilter;
	ofn.lpstrFile			= pstrFileName;
	ofn.lpstrFileTitle		= pstrTitleName;
	ofn.Flags				= OFN_HIDEREADONLY | OFN_CREATEPROMPT;

	return GetOpenFileName(&ofn);
}


BOOL PopFileSaveDlg(HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName)
{
	static char szFilter[] = "PGM files (*.pgm)\0*.pgm\0" \
							 "All Files (*.*)\0*.*\0\0";

	ofn.hwndOwner			= hwnd;
	ofn.lpstrFilter			= szFilter;
	ofn.lpstrFile			= pstrFileName;
	ofn.lpstrFileTitle		= pstrTitleName;
	ofn.lpstrDefExt			= "pgm";
	ofn.Flags				= OFN_OVERWRITEPROMPT;

	return GetSaveFileName(&ofn);
}


long FileLength(FILE *file)
{
	int iCurrentPos, iFileLength;

	iCurrentPos = ftell(file);
	fseek(file, 0, SEEK_END);
	iFileLength = ftell(file);
	fseek(file, iCurrentPos, SEEK_SET);

	return iFileLength;
}


BOOL PopFileRead(HWND hwndOutput, PSTR pstrFileName, Model *model)
{
	FILE	*fp;
	int		iLength;
	PSTR	pstrBuffer;
	BOOL	rets;

	rets = FALSE;
	fp = fopen(pstrFileName, "rt");
	if (fp == NULL)
		goto CLEANUP;

	iLength = FileLength(fp);
	switch (ofn.nFilterIndex)
	{
	case 1:
		rets = ReadDXFile(fp, model);
		break;
	}

CLEANUP:
	if (fp)
		fclose(fp);
	return rets;
}


BOOL ReadDXFile(FILE *fp, Model *model)
{	
	BOOL	rets;
	int		i, fileLen, filePos;
	int		*vList;
	char	temp[256], tmp2[256];
	DXObject	dx;
	Vertex	*vertex;
	Face	*face;

	rets = FALSE;
	fileLen = FileLength(fp);
	fseek(fp, 0, SEEK_SET);

	//read in VERTICES
	do
	{
		fgets(temp, 256, fp);
		if (strstr(temp, "object") != NULL)
			break;
		filePos = ftell(fp);
		if (filePos >= fileLen)
			goto CLEANUP;
	} while (TRUE);

	//Read in # of line segments
	//object 1 class array type float rank 1 shape 3 items 212 data follows
	sscanf(temp, "%s%d%s%s%s%s%s%d%s%d%s%d%s", tmp2, &dx.id, tmp2, tmp2, tmp2, dx.type,
			tmp2, &dx.rank, tmp2, &dx.shape, tmp2, &dx.items, tmp2);
	model->SetNumVertex(dx.items);
	vertex = model->GetVertex();
	for (i = 0; i < dx.items; i++)
	{
		fgets(temp, 256, fp);
		if (temp[0] == 0)
			break;
		if (strstr(temp, "object") != NULL)
			break;
		sscanf(temp, "%f%f%f", &vertex[i].v[0], &vertex[i].v[1], &vertex[i].v[2]); 
		filePos = ftell(fp);
		if (filePos >= fileLen)
			break;
	}
	
	//read in NORMALS
	do
	{
		if (strstr(temp, "object") != NULL)
			break;
		fgets(temp, 256, fp);
		filePos = ftell(fp);
		if (filePos >= fileLen)
			goto CLEANUP;
	} while (TRUE);

	//Read in # of line segments
	//object 1 class array type float rank 1 shape 3 items 212 data follows
	sscanf(temp, "%s%d%s%s%s%s%s%d%s%d%s%d%s", tmp2, &dx.id, tmp2, tmp2, tmp2, dx.type,
			tmp2, &dx.rank, tmp2, &dx.shape, tmp2, &dx.items, tmp2);
	model->SetNumVertexNormal(dx.items);
	vertex = model->GetVertexNormal();
	for (i = 0; i < dx.items; i++)
	{
		fgets(temp, 256, fp);
		if (temp[0] == 0)
			break;
		if (strstr(temp, "object") != NULL)
			break;
		sscanf(temp, "%f%f%f", &vertex[i].v[0], &vertex[i].v[1], &vertex[i].v[2]); 
		filePos = ftell(fp);
		if (filePos >= fileLen)
			break;
	}

	//read in FACES
	do
	{
		if (strstr(temp, "object") != NULL)
			break;
		fgets(temp, 256, fp);
		filePos = ftell(fp);
		if (filePos >= fileLen)
			goto CLEANUP;
	} while (TRUE);

	//Read in # of line segments
	//object 1 class array type float rank 1 shape 3 items 212 data follows
	sscanf(temp, "%s%d%s%s%s%s%s%d%s%d%s%d%s", tmp2, &dx.id, tmp2, tmp2, tmp2, dx.type,
			tmp2, &dx.rank, tmp2, &dx.shape, tmp2, &dx.items, tmp2);
	model->SetNumFace(dx.items);
	face = model->GetFace();
	for (i = 0; i < dx.items; i++)
	{
		fgets(temp, 256, fp);
		if (temp[0] == 0)
			break;
		if (strstr(temp, "object") != NULL)
			break;
		vList = face[i].GetVertexList();
		sscanf(temp, "%d%d%d", &vList[0], &vList[1], &vList[2]); 
		filePos = ftell(fp);
		if (filePos >= fileLen)
			break;
	}

//	model->GLCompileList();
	rets = TRUE;
CLEANUP:
	return rets;
}