//glcode.cpp

#include <windows.h>            // Window defines
#include <gl\gl.h>              // OpenGL
#include <gl\glu.h>             // GLU library
#include <math.h>				// Define for sqrt
#include "glcode.h"
#include "OutputView.h"
#include "Polygon.h"

#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)

#define HEAD_LIST       1
#define SHAFT_LIST      2
#define THREAD_LIST     3
#define BOLT_LIST       4
// Color Palette handle
//extern HPALETTE hPalette;

// Application name and instance storeage
//static LPCTSTR lpszAppName = "RenderBW";
//static HINSTANCE hInstance;

// Rotation amounts
static GLfloat xRot = 0.0f;
static GLfloat yRot = 0.0f;
static GLfloat xPos = 0.0f;
static GLfloat yPos = 0.0f;
static GLfloat nRange = 60.0f;


GLsizei lastHeight = 0, newHeight = 2;
GLsizei lastWidth = 0, newWidth = 2;

extern HWND	hOutput;
extern GLfloat *zbuffer;
extern GLfloat *pmcData;
extern GLfloat *gaussianData;
extern GLfloat *laplacianData;
extern GLfloat *outData[3];
extern GLubyte *outputBuf;
extern BOOL	isSceneDirty;
extern BOOL updateOutput;

extern Model	*curModel;

extern BOOL	processOutput;
extern BOOL resizeUpdate;
extern BOOL isButtonDown;
BOOL	doneDraw = FALSE;
// Set Pixel Format function - forward declaration
void SetDCPixelFormat(HDC hDC);



// Reduces a normal vector specified as a set of three coordinates,
// to a unit normal vector of length one.
void ReduceToUnit(float vector[3])
{
	float length;
	
	// Calculate the length of the vector		
	length = (float)sqrt((vector[0]*vector[0]) + 
		(vector[1]*vector[1]) +
		(vector[2]*vector[2]));
	
	// Keep the program from blowing up by providing an exceptable
	// value for vectors that may calculated too close to zero.
	if(length == 0.0f)
		length = 1.0f;
	
	// Dividing each element by the length will result in a
	// unit normal vector.
	vector[0] /= length;
	vector[1] /= length;
	vector[2] /= length;
}


// Points p1, p2, & p3 specified in counter clock-wise order
void calcNormal(float v[3][3], float out[3])
{
	float v1[3],v2[3];
	static const int x = 0;
	static const int y = 1;
	static const int z = 2;
	
	// Calculate two vectors from the three points
	v1[x] = v[0][x] - v[1][x];
	v1[y] = v[0][y] - v[1][y];
	v1[z] = v[0][z] - v[1][z];
	
	v2[x] = v[1][x] - v[2][x];
	v2[y] = v[1][y] - v[2][y];
	v2[z] = v[1][z] - v[2][z];
	
	// Take the cross product of the two vectors to get
	// the normal vector which will be stored in out
	out[x] = v1[y]*v2[z] - v1[z]*v2[y];
	out[y] = v1[z]*v2[x] - v1[x]*v2[z];
	out[z] = v1[x]*v2[y] - v1[y]*v2[x];
	
	// Normalize the vector (shorten length to one)
	ReduceToUnit(out);
}



// Change viewing volume and viewport.  Called when window is resized
void GLResize(GLsizei w, GLsizei h)
{
	GLsizei oldw, oldh, dataSize;
	GLfloat	fAspect;
	// Prevent a divide by zero
	if(h == 0)
		h = 1;
	newWidth = w;
	newHeight = h;
	oldw = lastWidth;
	oldh = lastHeight;
	resizeUpdate = TRUE;
	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);
	
	// Reset coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	fAspect = (GLfloat)w/(GLfloat)h;
	if (TRUE)
		gluPerspective(nRange, fAspect, 1.0, 300.0);
	// Establish clipping volume (left, right, bottom, top, near, far)
	else
	{
		if (w <= h) 
			glOrtho (-nRange, nRange, -nRange*h/w, nRange*h/w, -nRange, nRange);
	    else 
			glOrtho (-nRange*w/h, nRange*w/h, -nRange, nRange, -nRange, nRange);
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	//allocate memory 
	if (lastWidth != newWidth || lastHeight != newHeight)
	{
		int dataSize;
		
		lastWidth = newWidth;
		lastHeight = newHeight;
		dataSize = lastWidth * lastHeight;
		if (zbuffer != NULL)
			delete zbuffer;
		zbuffer = new GLfloat[dataSize];
		if (pmcData != NULL)
			delete pmcData;
		pmcData = new GLfloat[dataSize];
		outData[0] = pmcData;
		if (gaussianData != NULL)
			delete gaussianData;
		gaussianData = new GLfloat[dataSize];
		outData[1] = gaussianData;
		if (laplacianData != NULL)
			delete laplacianData;
		laplacianData = new GLfloat[dataSize];
		outData[2] = laplacianData;
		if (outputBuf != NULL)
			delete outputBuf;
		outputBuf = new GLubyte[dataSize];
		doneDraw = FALSE;
		isSceneDirty = TRUE;
		updateOutput = TRUE;
	}	

	if (!isButtonDown)
		resizeUpdate = FALSE;
}



// Creates the head of the bolt
void RenderHead(void)
{
	float x,y,angle;                                // Calculated positions
	float height = 25.0f;                   // Thickness of the head
	float diameter = 30.0f;                 // Diameter of the head
	float normal[3],corners[4][3];  // Storeage of vertices and normals
	float step = (3.1415f/3.0f);    // step = 1/6th of a circle = hexagon
	
	
	// Set material color for head of screw
	glColor3f(0.0f, 0.0f, 0.7f);
	
	// clock-wise polygons face out, set for fans
	glFrontFace(GL_CW);
	
	// Begin a new triangle fan to cover the top
	glBegin(GL_TRIANGLE_FAN);
	
	// All the normals for the top of the bolt point straight up
	// the z axis.
	glNormal3f(0.0f, 0.0f, 1.0f);
	
	// Center of fan is at the origin
	glVertex3f(0.0f, 0.0f, 0.0f);
	
	// Divide the circle up into 6 sections and start dropping
	// points to specify the fan
	for(angle = 0.0f; angle < (2.0f*3.1415f); angle += step)
	{
		// Calculate x and y position of the next vertex
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// Specify the next vertex for the triangle fan
		glVertex3f(x, y, 0.0f);
	}
	
	// Last vertex closes the fan
	glVertex3f(0.0f, diameter, 0.0f);
	
	// Done drawing the fan that covers the bottom
	glEnd();
	
	// Now draw the bottom of the bolt head. Now switch to
	// clock-wise polygons faceing out.
	glFrontFace(GL_CCW);                    
	
	// Begin a new triangle fan to cover the bottom
	glBegin(GL_TRIANGLE_FAN);
	
	// Normal for bottom points straight down the negative z axis
	glNormal3f(0.0f, 0.0f, -1.0f);
	
	// Center of fan is at the origin
	glVertex3f(0.0f, 0.0f, -height);
	
	// Divide the circle up into 6 sections and start dropping
	// points to specify the fan
	for(angle = 0.0f; angle < (2.0f*3.1415f); angle += step)
	{
		// Calculate x and y position of the next vertex
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// Specify the next vertex for the triangle fan
		glVertex3f(x, y, -height);
	}
	
	// Last vertex, used to close the fan
	glVertex3f(0.0f, diameter, -height);
	
	// Done drawing the fan that covers the bottom
	glEnd();
	
	
	// Build the sides out of triangles (two each). Each face
	// will consist of two triangles arranged to form a 
	// quadralateral
	glBegin(GL_TRIANGLES);
	
	// Go around and draw the sides
	for(angle = 0.0f; angle < (2.0f*3.1415f); angle += step)
	{
		// Calculate x and y position of the next hex point
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// start at bottom of head
		corners[0][0] = x;
		corners[0][1] = y;
		corners[0][2] = -height;
		
		// extrude to top of head
		corners[1][0] = x;
		corners[1][1] = y;
		corners[1][2] = 0.0f;
		
		// Calculate the next hex point
		x = diameter*(float)sin(angle+step);
		y = diameter*(float)cos(angle+step);
		
		// Make sure we aren't done before proceeding
		if(angle+step < 3.1415*2.0)
		{
			// If we are done, just close the fan at a 
			// known coordinate.
			corners[2][0] = x;
			corners[2][1] = y;
			corners[2][2] = 0.0f;
			
			corners[3][0] = x;
			corners[3][1] = y;
			corners[3][2] = -height;
		}
		else
		{
			// We aren't done, the points at the top and bottom
			// of the head.
			corners[2][0] = 0.0f;
			corners[2][1] = diameter;
			corners[2][2] = 0.0f;
			
			corners[3][0] = 0.0f;
			corners[3][1] = diameter;
			corners[3][2] = -height;
		}
		
		// The normal vectors for the entire face will
		// all point the same direction
		calcNormal(corners, normal);
		glNormal3fv(normal);
		
		// Specify each triangle seperately to lie next
		// to each other.
		glVertex3fv(corners[0]);
		glVertex3fv(corners[1]);
		glVertex3fv(corners[2]);
		
		glVertex3fv(corners[0]);
		glVertex3fv(corners[2]);
		glVertex3fv(corners[3]);
	}
	
	glEnd();
}



// Creates the shaft of the bolt as a cylinder with one end
// closed.
void RenderShaft(void)
{
	float x,y,angle;                        // Used to calculate cylinder wall
	float height = 75.0f;           // Height of the cylinder
	float diameter = 20.0f;         // Diameter of the cylinder
	float normal[3],corners[4][3];  // Storeage for vertices calculations
	float step = (3.1415f/50.0f);   // Approximate the cylinder wall with
	// 100 flat segments.
	
	
	// Set material color for head of screw
	glColor3f(0.0f, 0.0f, 0.7f);
	
	// counter clock-wise polygons face out (the default for triangles)
	glFrontFace(GL_CCW);
	
	
	// First assemble the wall as 100 quadrilaterals formed by
	// placing adjoining triangles together
	glBegin(GL_TRIANGLES);
	
	// Go around and draw the sides
	for(angle = 0.0f; angle < (2.0f*3.1415f); angle += step)
	{
		// Calculate x and y position of the next vertex
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// Get the coordinate for this point and extrude the 
		// length of the cylinder.
		corners[0][0] = x;
		corners[0][1] = y;
		corners[0][2] = -height;
		
		corners[1][0] = x;
		corners[1][1] = y;
		corners[1][2] = 0.0f;
		
		// Get the next point and do the same
		x = diameter*(float)sin(angle+step);
		y = diameter*(float)cos(angle+step);
		
		// If finished, use know starting point to close the surface
		if(angle+step < 3.1415*2.0) // Not Finished
		{
			corners[2][0] = x;
			corners[2][1] = y;
			corners[2][2] = 0.0f;
			
			corners[3][0] = x;
			corners[3][1] = y;
			corners[3][2] = -height;
		}
		else
		{
			// Finished, use the staring point
			corners[2][0] = 0.0f;
			corners[2][1] = diameter;
			corners[2][2] = 0.0f;
			
			corners[3][0] = 0.0f;
			corners[3][1] = diameter;
			corners[3][2] = -height;
		}
		
		// Instead of using real normal to actual flat section
		// Use what the normal would be if the surface was really
		// curved. Since the cylinder goes up the Z axis, the normal 
		// points from the Z axis out directly through each vertex.
		// Therefore we can use the vertex as the normal, as long as
		// we reduce it to unit length first.
		
		// First Triangle ////////////////////////////////////////
		// Fill the normal vector with the coordinate points
		normal[0] = corners[0][0];
		normal[1] = corners[0][1];
		normal[2] = corners[0][2];
		
		// Reduce to length of one and specify for this point
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[0]);
		
		// Get vertex, caluclate unit normal and go
		normal[0] = corners[1][0];
		normal[1] = corners[1][1];
		normal[2] = corners[1][2];
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[1]);
		
		// Get vertex, caluclate unit normal and go
		normal[0] = corners[2][0];
		normal[1] = corners[2][1];
		normal[2] = corners[2][2];
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[2]);
		
		
		// Second Triangle ////////////////////////////////////////
		
		// Get vertex, caluclate unit normal and go
		normal[0] = corners[2][0];
		normal[1] = corners[2][1];
		normal[2] = corners[2][2];
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[2]);
		
		// Get vertex, caluclate unit normal and go
		normal[0] = corners[3][0];
		normal[1] = corners[3][1];
		normal[2] = corners[3][2];
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[3]);
		
		// Get vertex, caluclate unit normal and go
		normal[0] = corners[0][0];
		normal[1] = corners[0][1];
		normal[2] = corners[0][2];
		ReduceToUnit(normal);
		glNormal3fv(normal);
		glVertex3fv(corners[0]);
	}
	
	glEnd();        // Done with cylinder sides
	
	// Begin a new triangle fan to cover the bottom
	glBegin(GL_TRIANGLE_FAN);
	
	// Normal points down the Z axis
	glNormal3f(0.0f, 0.0f, -1.0f);
	
	// Center of fan is at the origin
	glVertex3f(0.0f, 0.0f, -height);
	
	// Spin around matching step size of cylinder wall
	for(angle = 0.0f; angle < (2.0f*3.1415f); angle += step)
	{
		// Calculate x and y position of the next vertex
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// Specify the next vertex for the triangle fan
		glVertex3f(x, y, -height);
	}
	
	// Close the fan
	glVertex3f(0.0f, diameter, -height);
	glEnd();
}
	

// Creates the head of the bolt
void RenderThread(void)
{
	float x,y,z,angle;                              // Calculate coordinates and step angle
	float height = 75.0f;                   // Height of the threading
	float diameter = 20.0f;                 // Diameter of the threading
	float normal[3],corners[4][3];  // Storeage for calculate normal and corners
	float step = (3.1415f/32.0f);   // one revolution
	float revolutions = 7.0f;               // How many time around the shaft
	float threadWidth = 2.0f;               // How wide is the thread
	float threadThick = 3.0f;               // How thick is the thread
	float zstep = .125f;                    // How much does the thread move up
	// the Z axis each time a new segment
	// is drawn.
	
	// 360 degrees in radians
#define PI2 (2.0f*3.1415f)
	
	// Set material color for head of screw
	glColor3f(0.0f, 0.0f, 0.4f);    
	
	
	z = -height+2;  // Starting spot almost to the end
	
	// Go around and draw the sides until finished spinning up
	for(angle = 0.0f; angle < PI2*revolutions; angle += step)
	{
		// Calculate x and y position of the next vertex
		x = diameter*(float)sin(angle);
		y = diameter*(float)cos(angle);
		
		// Store the next vertex next to the shaft
		corners[0][0] = x;
		corners[0][1] = y;
		corners[0][2] = z;
		
		// Calculate the position away from the shaft
		x = (diameter+threadWidth)*(float)sin(angle);
		y = (diameter+threadWidth)*(float)cos(angle);
		
		corners[1][0] = x;
		corners[1][1] = y;
		corners[1][2] = z;
		
		// Calculate the next position away from the shaft
		x = (diameter+threadWidth)*(float)sin(angle+step);
		y = (diameter+threadWidth)*(float)cos(angle+step);
		
		corners[2][0] = x;
		corners[2][1] = y;
		corners[2][2] = z + zstep;
		
		// Calculate the next position along the shaft
		x = (diameter)*(float)sin(angle+step);
		y = (diameter)*(float)cos(angle+step);
		
		corners[3][0] = x;
		corners[3][1] = y;
		corners[3][2] = z+ zstep;
		
		
		// We'll be using triangels, so make
		// counter clock-wise polygons face out
		glFrontFace(GL_CCW);
		glBegin(GL_TRIANGLES);  // Start the top section of thread
		
		// Calculate the normal for this segment
		calcNormal(corners, normal);
		glNormal3fv(normal);
		
		// Draw two triangles to cover area
		glVertex3fv(corners[0]);
		glVertex3fv(corners[1]);
		glVertex3fv(corners[2]);
		
		glVertex3fv(corners[2]);
		glVertex3fv(corners[3]);
		glVertex3fv(corners[0]);
		
		glEnd();
		
		
		// Move the edge along the shaft slightly up the z axis
		// to represent the bottom of the thread
		corners[0][2] += threadThick;
		corners[3][2] += threadThick;
		
		// Recalculate the normal since points have changed, this
		// time it points in the opposite direction, so reverse it
		calcNormal(corners, normal);
		normal[0] = -normal[0];
		normal[1] = -normal[1];
		normal[2] = -normal[2];
		
		// Switch to clock-wise facing out for underside of the
		// thread.
		glFrontFace(GL_CW);
		
		// Draw the two triangles
		glBegin(GL_TRIANGLES);
		glNormal3fv(normal);
		
		glVertex3fv(corners[0]);
		glVertex3fv(corners[1]);
		glVertex3fv(corners[2]);
		
		glVertex3fv(corners[2]);
		glVertex3fv(corners[3]);
		glVertex3fv(corners[0]);
		
		glEnd();
		
		// Creep up the Z axis
		z += zstep;
	}
}

	
// This function does any needed initialization on the rendering
// context.  Here it sets up and initializes the lighting for
// the scene.
void GLSetupRC()
{
	// Light values and coordinates
	GLfloat  ambientLight[] = {0.4f, 0.4f, 0.4f, 1.0f };
	GLfloat  diffuseLight[] = {0.7f, 0.7f, 0.7f, 1.0f };
	GLfloat  specular[] = { 0.9f, 0.9f, 0.9f, 1.0f};
	GLfloat  lightPos[] = { -50.0f, 200.0f, 200.0f, 1.0f };
	GLfloat  specref[] =  { 0.6f, 0.6f, 0.6f, 1.0f };
	
	
	glEnable(GL_DEPTH_TEST);        // Hidden surface removal
	glEnable(GL_CULL_FACE);         // Do not calculate inside of solid object
	
	// Enable lighting
	glEnable(GL_LIGHTING);
	
	// Setup light 0
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,ambientLight);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambientLight);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuseLight);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);
	
	// Position and turn on the light
	glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
	glEnable(GL_LIGHT0);
	
	// Enable color tracking
	glEnable(GL_COLOR_MATERIAL);
	
	// Set Material properties to follow glColor values
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	
	// All materials hereafter have full specular reflectivity
	// with a moderate shine
	glMaterialfv(GL_FRONT, GL_SPECULAR,specref);
	glMateriali(GL_FRONT,GL_SHININESS,64);
	
	// Black background
	glClearColor(0.9f, 0.9f, 0.9f, 1.0f );
	
	if (curModel != NULL)
		curModel->GLCompileList();
	// Create display list for Bolt head
	glNewList(HEAD_LIST,GL_COMPILE);
	RenderHead();
	glEndList();
	
	// Create display list for shaft
	glNewList(SHAFT_LIST,GL_COMPILE);
	RenderShaft();
	glEndList();
	
	// Create display list for thread
	glNewList(THREAD_LIST,GL_COMPILE);
	RenderThread();
	glEndList();
	
	
	// Create nested display list for entire bolt
	glNewList(BOLT_LIST,GL_COMPILE);
	
	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Make sure we have the correct matrix mode
	glMatrixMode(GL_MODELVIEW);
	
	// Translate and render the head
	glTranslatef(0.0f, 0.0f, 55.0f);
	glCallList(HEAD_LIST);
	
	// Translate back some and render the shaft and thread together
	glTranslatef(0.0f, 0.0f, -15.0f);
	glCallList(SHAFT_LIST);
	glCallList(THREAD_LIST);
	
	// Translate back again for next pass
	glTranslatef(0.0f, 0.0f, -40.0f);
	
	// End Bolt list
	glEndList();
}



// Called to draw the entire bolt
void GLRenderScene(void)
{
	// Save the matrix state and do the rotations
	glPushMatrix();
	glTranslatef(xPos, yPos, -150.0f);
	glRotatef(xRot, 1.0f, 0.0f, 0.0f);
	glRotatef(yRot, 0.0f, 1.0f, 0.0f);
	
	if (curModel == NULL)
		glCallList(BOLT_LIST);
	else glCallList(curModel->GetListID());
	glPopMatrix();
	// Flush drawing commands
	glFlush();

	glReadPixels(0, 0, lastWidth, lastHeight, GL_DEPTH_COMPONENT, GL_FLOAT, (void*) zbuffer);

	if (!doneDraw)
	{
		doneDraw = TRUE;
		//resizeUpdate = FALSE;
		isSceneDirty = TRUE;
		updateOutput = TRUE;
	}
/*	if (isSceneDirty)
	{
		CalcMeanCurv(zbuffer);
		isSceneDirty = FALSE;
	}
	if (updateOutput)
	{
		ProcessOutput();
		updateOutput = FALSE;
	}*/
	//InvalidateRect(hOutput, NULL, FALSE);
}
	

void KeyUp()
{
	if (GetKeyState(VK_SHIFT) < 0)
		yPos += 5.0f;
	else
		xRot -= 5.0f;
	CheckRotateAngle();
}

void KeyDown()
{
	if (GetKeyState(VK_SHIFT) < 0)
		yPos -= 5.0f;
	else 
		xRot += 5.0f;
	CheckRotateAngle();
}

void KeyLeft()
{
	if (GetKeyState(VK_SHIFT) < 0)
		xPos -= 5.0f;
	else 
		yRot -= 5.0f;
	CheckRotateAngle();
}

void KeyRight()
{
	if (GetKeyState(VK_SHIFT) < 0)
		xPos += 5.0f;
	else 
		yRot += 5.0f;
	CheckRotateAngle();
}

void CheckRotateAngle()
{
	xRot = (const int)xRot % 360;
	yRot = (const int)yRot % 360;
}

void ZoomIn()
{
	nRange--;
	if (nRange < 10)
		nRange =10;
//	else GLResize(lastWidth, lastHeight);
}

void ZoomOut()
{
	nRange++;
	if (nRange > 200)
		nRange = 200;
//	else GLResize(lastWidth, lastHeight);
}
