#include <windows.h>
#include "LinkList.h"

//Class for ListItem, used in LinkList class
ListItem::ListItem(void* data)
{
	m_data = data;
	m_prev = NULL;
	m_next = NULL;
}


ListItem::~ListItem()
{
}

void* ListItem::GetItem()
{
	return m_data;
}

ListItem* ListItem::Prev()
{
	return m_prev;
}

ListItem* ListItem::Next()
{
	return m_next;
}

void ListItem::SetPrev(ListItem *prev)
{
	m_prev = prev;
}

void ListItem::SetNext(ListItem *next)
{
	m_next = next;
}

//Class for LinkList
LinkList::LinkList()
{
	m_head = NULL;
	m_tail = NULL;
	m_numItem = 0;
}


LinkList::~LinkList()
{
	EmptyList();
}

		
int	LinkList::GetNumItem()
{
	return m_numItem;
}


void LinkList::EmptyList()
{
	while (m_head != NULL) {
		DeleteItemAt(0);
	}
	m_head = NULL;
	m_tail = NULL;
	m_numItem = 0;
}


void LinkList::InsertItem(ListItem *item)
{
	InsertItemAt(item, m_numItem);
}

void LinkList::InsertItemData(void *data)
{
	ListItem *item;

	item = new ListItem(data);
	if (item == NULL)
		return;
	InsertItemAt(item, m_numItem);
}


void LinkList::InsertItemAt(ListItem *item, int pos)
{
	int	i;
	ListItem *temp;

	if (item == NULL)
		return;

	if (m_head == NULL) {
		m_head = item;
		m_tail = item;
		m_numItem++;
	}
	else if (pos == 0) {
		m_head->SetPrev(item);
		item->SetNext(m_head);
		m_head = item;
		m_numItem++;
	}
	else if (pos >= m_numItem) {
		m_tail->SetNext(item);
		item->SetPrev(m_tail);
		m_tail = item;
		m_numItem++;
	}
	else {
		temp = m_head;
		i = 0;
		while (i < pos && i < m_numItem) {
			temp = temp->Next();
			i++;
		}
		item->SetNext(temp);
		item->SetPrev(temp->Prev());
		temp->Prev()->SetNext(item);
		temp->SetPrev(item);
		m_numItem++;
	}	
}

void LinkList::DeleteItem(void *data)
{
	int pos;

	pos = GetPosOfItem(data);
	DeleteItemAt(pos);
}

void LinkList::DeleteItemAt(int pos)
{
	int	i;
	ListItem	*temp;

	if (m_numItem == 0 || pos < 0)
		return;

	if (m_numItem == 1) {	//If there's only 1 item
		if (m_head)
			delete m_head;
		m_head = NULL;
		m_tail = NULL;
		m_numItem = 0;
	}
	else if (pos == 0) {	//deleting the head
		temp = m_head->Next();
		temp->SetPrev(NULL);
		if (m_head)
			delete m_head;
		m_head = temp;
		m_numItem--;
	}
	else if (pos >= m_numItem - 1) {	//deleting tail
		temp = m_tail->Prev();
		temp->SetNext(NULL);
		if (m_tail)
			delete m_tail;
		m_tail = temp;
		m_numItem--;
	}
	else {					//deleting item in the list
		temp = m_head;
		i = 0;
		while (i < pos && i < m_numItem) {
			temp = temp->Next();
			i++;
		}
		temp->Prev()->SetNext(temp->Next());
		temp->Next()->SetPrev(temp->Prev());
		if (temp)
			delete temp;
		temp = NULL;
		m_numItem--;
	}
}


ListItem* LinkList::GetItemAt(int pos)
{
	int	i;
	ListItem *temp;

	if (m_numItem == 0 || pos < 0)
		return NULL;

	temp = m_head;
	i = 0;
	while (i < pos && i < m_numItem) {
		temp = temp->Next();
		i++;
	}
	return temp;
}


void* LinkList::GetItemDataAt(int pos)
{
	ListItem *item;

	item = GetItemAt(pos);
	return item->GetItem();
}


int	LinkList::GetPosOfItem(void* data)
{
	int i;
	ListItem *temp;

	temp = m_head;
	i = 0;
	while (i < m_numItem && temp) {
		if (temp->GetItem() == data)
			break;
		temp = temp->Next();
		i++;
	}
	if (!temp)
		i = -1;
	return i;
}
