

#include <windows.h>            // Window defines
#include <math.h>
#include <gl\gl.h>              // OpenGL
#include <gl\glu.h>             // GLU library

#include "OutputView.h"
#include "RenderBW.h"
#include "glcode.h"
#include "DialogBox.h"

// Palette handle, created by main window
//extern "C"
//{
extern HGLRC hRC;
extern HDC hDC; 
	extern HPALETTE hPalette;
	extern HWND hwndMain;
	extern HWND hOutput;
	extern int viewMode;
	extern 	BOOL menuChecked[3];
	extern ThresholdStruct *pThreshold;
	extern BOOL	isSceneDirty;
	extern BOOL updateOutput;
	extern BOOL refreshOutput;
	extern BOOL	isButtonDown;
	// Permenant Rendering context
//}



extern GLsizei lastHeight, newHeight;
extern GLsizei lastWidth, newWidth;
//
GLfloat *zbuffer = NULL;
GLfloat *pmcData = NULL;
GLfloat *gaussianData = NULL;
GLfloat *laplacianData = NULL;
GLfloat *outData[3];
GLubyte *outputBuf = NULL;

BOOL	processOutput = FALSE;
BOOL	resizeUpdate = FALSE;
// Declaration for Window procedure
LRESULT CALLBACK OutputWndProc(   HWND    hwnd,
							   UINT    message,
							   WPARAM  wParam,
							   LPARAM  lParam);


void CreateOutputWindow(HWND hParent, HINSTANCE hInst)
{
	WNDCLASS        wc;                     // Windows class structure
	
	// Register Window style
	wc.style                = CS_HREDRAW | CS_VREDRAW  | CS_OWNDC;
	wc.lpfnWndProc          = (WNDPROC) OutputWndProc;
	wc.cbClsExtra           = 0;
	wc.cbWndExtra           = 0;
	wc.hInstance            = hInst;
	wc.hIcon                = NULL;
	wc.hCursor              = LoadCursor(NULL, IDC_ARROW);
	
	// No need for background brush for OpenGL window
	wc.hbrBackground        = NULL;
	
	wc.lpszMenuName         = NULL;
	wc.lpszClassName        = "OutputView";
	
	// Register the window class
	RegisterClass(&wc);
	
	
	// Create the main application window
	hOutput = CreateWindow(
		wc.lpszClassName,
		"Output",
		
		// OpenGL requires WS_CLIPCHILDREN and WS_CLIPSIBLINGS
		/*WS_CHILD |WS_OVERLAPPED*/WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		
		// Window position and size
		500, 30,
		250, 250,
		hParent,
		NULL,
		hInst,
		NULL);
	
	// Display the window
	ShowWindow(hOutput, SW_SHOW);
	UpdateWindow(hOutput);
}





// Window procedure, handles all messages for this program
LRESULT CALLBACK OutputWndProc( HWND    hwnd,
							   UINT    message,
							   WPARAM  wParam,
							   LPARAM  lParam)
{
	static HDC hDC1;                 // Private GDI Device context
	static HGLRC hRC1;             
	switch (message)
	{
		
		// Window creation, setup for OpenGL
	case WM_CREATE:
		// Store the device context
		hDC1 = GetDC(hwnd);
		
		// Select the pixel format
		SetDCPixelFormat(hDC1);
		
		// Create the rendering context and make it current
		hRC1 = wglCreateContext(hDC1);
		wglMakeCurrent(hDC1, hRC1);
		//GLSetupRC();
		
		SetTimer(hwnd,101,1000,NULL);
		
		break;
		
	case WM_TIMER:
		InvalidateRect(hwnd, NULL, FALSE);
		break;
		
		// Window is being destroyed, cleanup
	case WM_DESTROY:
		// Deselect the current rendering context and delete it
		wglMakeCurrent(hDC1, NULL);
		wglDeleteContext(hRC1);
		ReleaseDC(hwnd, hDC1);
		KillTimer(hwnd, 101);
		hOutput = NULL;
		break;
		
		// Window is resized.
	case WM_SIZE:
		// Call our function which modifies the clipping
		// volume and viewport
		//wglMakeCurrent(hDC, hRC1);
		//GLResize(LOWORD(lParam), HIWORD(lParam));
		//wglMakeCurrent(hDC, NULL);
		break;
		
		
		// The painting function.  This message sent by Windows
		// whenever the screen needs updating.
	case WM_PAINT:
		{
			int dataSize;
			RECT rect;
			
			if (resizeUpdate)
				break;
			if (refreshOutput)
			{
				int dataSize;
				RECT rect2;

				GetClientRect(hwndMain, &rect2);
				newWidth = rect2.right;
				newHeight = rect2.bottom;
				lastWidth = lastHeight = 1;
				isSceneDirty = TRUE;
				updateOutput = TRUE;
				wglMakeCurrent(hDC, hRC);
				GLSetupRC();
 				GLResize(newWidth, newHeight);
				GLRenderScene();
 				glReadPixels(0, 0, lastWidth, lastHeight, GL_DEPTH_COMPONENT, GL_FLOAT, (void*) zbuffer);
				wglMakeCurrent(hDC, NULL);
				refreshOutput = FALSE;
			}
			if (isSceneDirty)
			{
				CalcMeanCurv(zbuffer);
				isSceneDirty = FALSE;
			}
			if (updateOutput && zbuffer)
			{
				ProcessOutput();
				updateOutput = FALSE;
			}
			
			wglMakeCurrent(hDC1, hRC1);
			// Call OpenGL drawing code
			// Rotate and translate the coordinate system
			// Note this will be cumulative
			//glRotatef(5.0f, 0.0f, 1.0f, 0.0f);
			GetClientRect(hwnd, &rect);
			if (zbuffer)
				OutputView(&rect);
			
			// Bring image to front
			SwapBuffers(hDC1);
			
			wglMakeCurrent(hDC1, NULL);
			
			// Do not validate, forcing a continuous repaint
			ValidateRect(hwnd, &rect);
		}
		break;
		
		// Windows is telling the application that it may modify
		// the system palette.  This message in essance asks the
		// application for a new palette.
	case WM_QUERYNEWPALETTE:
		// If the palette was created.
		if(hPalette)
		{
			int nRet;
			
			// Selects the palette into the current device context
			SelectPalette(hDC1, hPalette, FALSE);
			
			// Map entries from the currently selected palette to
			// the system palette.  The return value is the number
			// of palette entries modified.
			nRet = RealizePalette(hDC1);
			
			// Validate window contents
			ValidateRect(hwnd,FALSE);
			
			return nRet;
		}
		break;
		
	case WM_ERASEBKGND:
		return FALSE;
		break;
		
		// This window may set the palette, even though it is not the
		// currently active window.
	case WM_PALETTECHANGED:
		// Don't do anything if the palette does not exist, or if
		// this is the window that changed the palette.
		if((hPalette != NULL) && ((HWND)wParam != hwnd))
		{
			// Select the palette into the device context
			SelectPalette(hDC1,hPalette,FALSE);
			
			// Map entries to system palette
			RealizePalette(hDC1);
			
			// Remap the current colors to the newly realized palette
			UpdateColors(hDC1);
			return 0;
		}
		break;
		
	default:   // Passes it on if unproccessed
		return (DefWindowProc(hwnd, message, wParam, lParam));
		
	}
	
    return (0L);
}


void OutputView(RECT *rect)
{
	GLint	xoffset, yoffset;
	GLint	xsize, ysize;
	GLfloat	xscale, yscale;

	glViewport(0, 0, rect->right, rect->bottom);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0, rect->right - 1.0, 0.0, rect->bottom - 1.0, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);
	
	xsize = rect->right;
	ysize = lastHeight * xsize / lastWidth;
	if (ysize > rect->bottom)
	{
		ysize = rect->bottom;
		xsize = lastWidth * ysize / lastHeight;
	}
	
	xscale = (float)xsize / lastWidth;
	yscale = (float)ysize / lastHeight;
	xoffset = (rect->right - xsize) * 0.5;
	yoffset = (rect->bottom - ysize) * 0.5;
	
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	glPixelZoom(xscale, yscale);
	glRasterPos2i(xoffset, yoffset);
	
	//if (viewMode == 0)
	//{
		//glDrawPixels(lastWidth, lastHeight, GL_LUMINANCE, GL_FLOAT, (void*)zbuffer);
	//}
	//else
	//{
	//glDrawPixels(lastWidth, lastHeight, GL_LUMINANCE, GL_FLOAT, (void*)pmcData);
	//glDrawPixels(lastWidth, lastHeight, GL_LUMINANCE, GL_FLOAT, (void*)gaussianData);
	//else
	glDrawPixels(lastWidth, lastHeight, GL_LUMINANCE, GL_UNSIGNED_BYTE, (void*)outputBuf);
	//}
	glFinish();
}


//Curvature calculations
void CalcMeanCurv(GLfloat *zbuf)
{
	GLfloat fx, fy;
	GLfloat	fxx, fyy, fxy;
	GLfloat tmp;

	int i, j, yoffset, offset;

	pThreshold->max[3] = 0;
	pThreshold->min[3] = 1.0;
	processOutput = TRUE;
	yoffset = 0;
	for (j = 1; j < lastHeight - 1; j++)
	{
		yoffset += lastWidth;
		for (i = 1; i < lastWidth - 1; i++)
		{
			offset = yoffset + i;
			fx = (zbuf[offset+1] - zbuf[offset-1]) / 2;
			fy = (zbuf[offset+lastWidth] - zbuf[offset-lastWidth]) /2;
			fxx = (zbuf[offset-1] - 2*zbuf[offset] + zbuf[offset+1]) / 4;
			fyy = (zbuf[offset-lastWidth] - 2 * zbuf[offset] + zbuf[offset+lastWidth]) / 4;
			fxy = (-zbuf[offset-1-lastWidth] + zbuf[offset+1-lastWidth])/4 +
				(zbuf[offset-1+lastWidth] - zbuf[offset+1+lastWidth])/4;
			tmp =	(fxx + fyy + fxx*fy*fy + fyy*fx*fx - 2*fx*fy*fxy) /
				(2 * pow( (1+fx*fx+ fy*fy), 1.5));
			if (i == 1 && j == 1)
				pThreshold->max[0] = pThreshold->min[0] = tmp;
			else
			{
				if (tmp > pThreshold->max[0])
					pThreshold->max[0] = tmp;
				else if (tmp < pThreshold->min[0])
					pThreshold->min[0] = tmp;
			}
			pmcData[offset] = tmp;
			//if (pos)
			//	pmcData[offset] = (tmp > threshold1 ? 0.0 : 1.0);
			//else
			//	pmcData[offset] = (tmp < threshold1 ? 0.0 : 1.0);

			tmp =	(fxx*fyy - fxy*fxy) / pow((1+fx*fx+fy*fy), 2);
			if (i == 1 && j == 1)
				pThreshold->max[1] = pThreshold->min[1] = tmp;
			else
			{
				if (tmp > pThreshold->max[1])
					pThreshold->max[1] = tmp;
				else if (tmp < pThreshold->min[1])
					pThreshold->min[1] = tmp;
			}
			gaussianData[offset] = tmp;

			tmp = - zbuf[offset+1] - zbuf[offset-1] - zbuf[offset+lastWidth] - zbuf[offset-lastWidth]
			 + 8 *zbuf[offset]
			-zbuf[offset-1-lastWidth] - zbuf[offset+1-lastWidth] - zbuf[offset-1+lastWidth] - zbuf[offset+1+lastWidth];
			
			if (i == 1 && j == 1)
				pThreshold->max[2] = pThreshold->min[2] = tmp;
			else
			{
				if (tmp > pThreshold->max[2])
					pThreshold->max[2] = tmp;
				else if (tmp < pThreshold->min[2])
					pThreshold->min[2] = tmp;
			}
			laplacianData[offset] = tmp;
		
			if (zbuf[offset] != 1.0)
			{
				if (zbuf[offset] > pThreshold->max[3])
					pThreshold->max[3] = zbuf[offset];
				else if (zbuf[offset] < pThreshold->min[3])
					pThreshold->min[3] = zbuf[offset];
			}

		}
	}
	for (i = 0; i < 4; i++)
	{
		pThreshold->median[i] = (pThreshold->max[i] + pThreshold->min[i]) / 2;
		pThreshold->range[i] = pThreshold->max[i] - pThreshold->min[i];
	}
}

void ProcessOutput(void)
{
	GLfloat	val[3][2];
	GLubyte *byte, val2;
	GLsizei	width, height;
	register int	i, x, y, hoffset, offset;
	
	width = lastWidth;
	height = lastHeight;

	processOutput = TRUE;
	memset(outputBuf, 255, sizeof(GLubyte)*lastWidth*lastHeight);
	for (i = 0; i < 3; i++)
	{
		val[i][0] = pow(pThreshold->curve[i],10) * pThreshold->max[i] / pow(RANGE,10);
		val[i][1] = pow(pThreshold->curve[i],10) * pThreshold->min[i] / pow(RANGE,10);
	}
	hoffset = -lastWidth;
	for (y = 0; y < lastHeight; y++)
	{
		hoffset += lastWidth;
		for (x = 0; x < lastWidth; x++)
		{
			offset = hoffset + x;
			byte = &outputBuf[offset];
			for (i = 0; i < 3; i++)
			{
				if (viewMode == 0)
				{
					if (zbuffer[offset] != 1.0)
					{
						val2 = (zbuffer[offset] - pThreshold->min[3]) / pThreshold->range[3] * 255;
						*byte &= val2;
					}
				}
				else
				{
					if (menuChecked[i])
					{
						if (pThreshold->blackwhite[i])
						{
							//if (pThreshold->curve[i] >= 0)
							//	val = pow(pThreshold->curve[i],2) * pThreshold->max[i] / pow(RANGE,2);
							//else if (pThreshold->curve[i] < 0)
							//	val = pow(pThreshold->curve[i],2) * pThreshold->min[i] / pow(RANGE,2);
							if (pThreshold->sign[i] & POSITIVE)
							{
								if (outData[i][offset] > val[i][0])
									*byte &= 0;
								//else byte |= 255;
							}
							if (pThreshold->sign[i] & NEGATIVE)
							{
								if (outData[i][offset] < val[i][1])
									*byte &= 0;
								//else byte |= 255;
							}
						}
						else
						{
							if (pThreshold->positive[i] & POSITIVE)
								val2 = 255 - (pow(outData[i][offset], 0.1) / pow(pThreshold->max[0], 0.1)) * 255;
							if (pThreshold->positive[i] & NEGATIVE)
								val2 = 255 - (pow(outData[i][offset], 0.1) / pow(pThreshold->min[0], 0.1)) * 255;
							*byte &= val2;
						}
					}
				}
			}
		}
	}
	updateOutput = FALSE;
	processOutput = FALSE;
}