#include <windows.h>
#include "gl\gl.h"
#include "Polygon.h"
#include "glcode.h"

Face::Face()
{
	m_numVertex = 3;
	m_vertexList = new int[3];
}

Face::Face(int num)
{
	m_numVertex = num;
	m_vertexList = new int[num];
}

Face::~Face()
{
	if (m_vertexList)
		delete[] m_vertexList;
}



//===============================

Model::Model()
{
	static int nextID = 10;

	m_glListID = nextID++;
	m_numVertex = 0;
	m_vertex = NULL;
	m_numVNormal = 0;
	m_vertexNormal = NULL;
	m_numFace = 0;
	m_face = NULL;
}

Model::~Model()
{
	if (m_vertex)
		delete[] m_vertex;
	if (m_vertexNormal)
		delete[] m_vertexNormal;
	if (m_face)
		delete[] m_face;
}

void Model::SetNumVertex(int num)
{
	m_numVertex = num;
	m_vertex = new Vertex[num];
}

void Model::SetNumVertexNormal(int num)
{
	m_numVNormal = num;
	m_vertexNormal = new Vertex[num];
}

void Model::SetNumFace(int num)
{
	m_numFace = num;
	m_face = new Face[num];
}

void Model::GLDraw()
{
	int i, vlist[3];
	float normal[3];
	float v[3][3];
	
	// Set material color
	glColor3f(0.0f, 0.0f, 0.4f);    
	glFrontFace(GL_CCW);
	glBegin(GL_TRIANGLES);
	for (i = 0 ; i < m_numFace ; i++)
	{
		vlist[0] = (m_face[i].GetVertexList())[0];
		vlist[1] = (m_face[i].GetVertexList())[1];
		vlist[2] = (m_face[i].GetVertexList())[2];
		v[0][0] = m_vertex[vlist[0]].v[0];
		v[0][1] = m_vertex[vlist[0]].v[1];
		v[0][2] = m_vertex[vlist[0]].v[2];
		v[1][0] = m_vertex[vlist[1]].v[0];
		v[1][1] = m_vertex[vlist[1]].v[1];
		v[1][2] = m_vertex[vlist[1]].v[2];
		v[2][0] = m_vertex[vlist[2]].v[0];
		v[2][1] = m_vertex[vlist[2]].v[1];
		v[2][2] = m_vertex[vlist[2]].v[2];
		
		if (m_vertexNormal == NULL)
		{
			calcNormal(v, normal);
			glNormal3fv(normal);
			glVertex3fv(v[0]);
			glVertex3fv(v[1]);
			glVertex3fv(v[2]);
		}
		else
		{
			glNormal3fv(m_vertexNormal[vlist[0]].v);
			glVertex3fv(v[0]);
			glNormal3fv(m_vertexNormal[vlist[1]].v);
			glVertex3fv(v[1]);
			glNormal3fv(m_vertexNormal[vlist[2]].v);
			glVertex3fv(v[2]);
		}

	}
	glEnd();
}

void Model::GLCompileList()
{
	// Create nested display list for entire bolt
	glNewList(m_glListID, GL_COMPILE);
	
	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Make sure we have the correct matrix mode
	glMatrixMode(GL_MODELVIEW);
	
	GLDraw();

	glEndList();

}