// ShinyJet.c
// Draws a crude approximation of an aircraft. Material properties are set
// to make the surface appear shiny

#include <windows.h>            // Window defines
#include <commdlg.h>
#include <stdlib.h>
#include <gl\gl.h>              // OpenGL
#include <gl\glu.h>             // GLU library
#include <math.h>				// Define for sqrt
#include <AFXRES.H>
#include "resource.h"           // About box resource identifiers.

#include "glcode.h"
#include "OutputView.h"
#include "RenderBW.h"
#include "Dialogbox.h"
#include "File_io.h"
#include "Polygon.h"
#include "pgm.h"

#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)

HGLRC hRC;               // Permenant Rendering context
HDC hDC;                 // Private GDI Device context

// Color Palette handle
HPALETTE hPalette = NULL;
HWND hwndMain;
HWND hOutput;
HWND hDlgModeless;
ThresholdStruct *pThreshold = NULL;
int		viewMode = 0;
BOOL	menuChecked[3];
BOOL	isSceneDirty = TRUE;
BOOL	updateOutput = TRUE;
BOOL	refreshOutput = FALSE;
BOOL	isButtonDown = FALSE;
Model	*curModel = NULL;


extern GLfloat *zbuffer;
extern GLfloat *pmcData;
extern GLfloat *gaussianData;
extern GLfloat *laplacianData;
extern GLubyte *outputBuf;

// Application name and instance storeage
static LPCTSTR szAppName = "RenderBW";
static HINSTANCE hInstance;


extern GLsizei lastHeight;
extern GLsizei lastWidth;

extern BOOL	processOutput;

// Declaration for Window procedure
LRESULT CALLBACK WndProc(   HWND    hwnd,
						 UINT    message,
						 WPARAM  wParam,
						 LPARAM  lParam);

// Dialog procedure for about box
//BOOL APIENTRY AboutDlgProc (HWND hDlg, UINT message, UINT wParam, LONG lParam);
BOOL CALLBACK AboutDlgProc (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);



// Select the pixel format for a given device context
void SetDCPixelFormat(HDC hDC)
{
	int nPixelFormat;
	
	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),  // Size of this structure
			1,                                                              // Version of this structure    
			PFD_DRAW_TO_WINDOW |                    // Draw to Window (not to bitmap)
			PFD_SUPPORT_OPENGL |					// Support OpenGL calls in window
			PFD_DOUBLEBUFFER,                       // Double buffered
			PFD_TYPE_RGBA,                          // RGBA Color mode
			24,                                     // Want 24bit color 
			0,0,0,0,0,0,                            // Not used to select mode
			0,0,                                    // Not used to select mode
			0,0,0,0,0,                              // Not used to select mode
			32,                                     // Size of depth buffer
			0,                                      // Not used to select mode
			0,                                      // Not used to select mode
			PFD_MAIN_PLANE,                         // Draw in main plane
			0,                                      // Not used to select mode
			0,0,0 };                                // Not used to select mode
		
		// Choose a pixel format that best matches that described in pfd
		nPixelFormat = ChoosePixelFormat(hDC, &pfd);
		
		// Set the pixel format for the device context
		SetPixelFormat(hDC, nPixelFormat, &pfd);
}



// If necessary, creates a 3-3-2 palette for the device context listed.
HPALETTE GetOpenGLPalette(HDC hDC)
{
	HPALETTE hRetPal = NULL;	// Handle to palette to be created
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;						// Counting variable
	BYTE RedRange,GreenRange,BlueRange;
	// Range for each color entry (7,7,and 3)
	
	
	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
	
	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return NULL
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return NULL;
	
	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;	
	
	// Allocate space for a logical palette structure plus all the palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +nColors*sizeof(PALETTEENTRY));
	
	// Fill in palette header 
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; // table size
	
	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.  
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;
	
	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
	{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);
		
		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) & GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);
		
		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);
		
		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
	}
	
	
	// Create the palette
	hRetPal = CreatePalette(pPal);
	
	// Go ahead and select and realize the palette for this device context
	SelectPalette(hDC,hRetPal,FALSE);
	RealizePalette(hDC);
	
	// Free the memory used for the logical palette structure
	free(pPal);
	
	// Return the handle to the new palette
	return hRetPal;
}


// Entry point of all Windows programs
int WINAPI WinMain(   HINSTANCE       hInst,
					 HINSTANCE       hPrevInstance,
					 LPSTR           lpCmdLine,
					 int                     nCmdShow)
{
	MSG                     msg;            // Windows message structure
	WNDCLASS        wc;                     // Windows class structure
	HWND            hwnd;           // Storeage for window handle
	
	hInstance = hInst;
	
	// Register Window style
	wc.style                        = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc          = (WNDPROC) WndProc;
	wc.cbClsExtra           = 0;
	wc.cbWndExtra           = 0;
	wc.hInstance            = hInstance;
	wc.hIcon                        = NULL;
	wc.hCursor                      = LoadCursor(NULL, IDC_ARROW);
	
	// No need for background brush for OpenGL window
	wc.hbrBackground        = NULL;         
	
	wc.lpszMenuName         = MAKEINTRESOURCE(IDR_MENU);
	wc.lpszClassName        = szAppName;
	
	// Register the window class
	if(RegisterClass(&wc) == 0)
		return FALSE;
	
	
	// Create the main application window
	hwnd = CreateWindow(
		szAppName,
		szAppName,
		
		// OpenGL requires WS_CLIPCHILDREN and WS_CLIPSIBLINGS
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		
		// Window position and size
		50, 50,
		400, 400,
		NULL,
		NULL,
		hInstance,
		NULL);
	
	// If window was not created, quit
	if(hwnd == NULL)
		return FALSE;
	
	hwndMain = hwnd;
	// Display the window
	ShowWindow(hwnd,SW_SHOW);
	UpdateWindow(hwnd);
	
	// Process application messages until the application closes
	while( GetMessage(&msg, NULL, 0, 0))
	{
		if (hDlgModeless == 0 || !IsDialogMessage(hDlgModeless, &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	
	return msg.wParam;
}



//Common dialogs
void DoCaption(HWND hwnd, char *szTitleName)
{
	char szCaption[64+_MAX_FNAME + _MAX_EXT];

	wsprintf(szCaption, "%s - %s", szAppName, szTitleName[0] ? szTitleName : UNTITLED);
	SetWindowText(hwnd, szCaption);
}

void OkMessage(HWND hwnd, char *szMessage, char *szTitleName)
{
	char szBuffer[64 + _MAX_FNAME + _MAX_EXT];

	wsprintf(szBuffer, szMessage, szTitleName[0] ? szTitleName : UNTITLED);
	MessageBox(hwnd, szBuffer, szAppName, MB_OK | MB_ICONEXCLAMATION);
}


// Window procedure, handles all messages for this program
LRESULT CALLBACK WndProc(       HWND    hwnd,
						 UINT    message,
						 WPARAM  wParam,
						 LPARAM  lParam)
{
	static char szFileName[_MAX_PATH];
	static char szTitleName[_MAX_FNAME + _MAX_EXT];
	
	switch (message)
	{
		// Window creation, setup for OpenGL
	case WM_CREATE:
		// Store the device context
		hDC = GetDC(hwnd);              
		
		// Select the pixel format
		SetDCPixelFormat(hDC);          
		
		// Create palette if needed
		hPalette = GetOpenGLPalette(hDC);
		
		// Create the rendering context and make it current
		hRC = wglCreateContext(hDC);
		wglMakeCurrent(hDC, hRC);
		GLSetupRC();
 		wglMakeCurrent(hDC, NULL);
  
		pThreshold = new ThresholdStruct;
		if (pThreshold == NULL)
			break;
		
		int i;
		for (i = 0; i < 3; i++)
		{
			menuChecked[i] = TRUE;
			pThreshold->blackwhite[i] = TRUE;
			pThreshold->positive[i] = TRUE;
			pThreshold->curve[i] = 0.0;
			pThreshold->sign[i] = 1;
		}

		//create output view
		CreateOutputWindow(hwnd,hInstance);
		CheckMenuItem(GetMenu(hwnd), ID_VIEW_ZBUFFER, MF_CHECKED);		
		CheckMenuItem(GetMenu(hwnd), ID_VIEW_PMC, MF_CHECKED);		
		CheckMenuItem(GetMenu(hwnd), ID_VIEW_GAUSSIAN, MF_CHECKED);		
		CheckMenuItem(GetMenu(hwnd), ID_VIEW_LAPLACIAN, MF_CHECKED);	
		
		PopFileInitialize(hwnd);
		break;
		
		// Window is being destroyed, cleanup
	case WM_DESTROY:
		if (zbuffer != NULL)
			delete zbuffer;
		if (pmcData != NULL)
			delete pmcData;
		if (gaussianData != NULL)
			delete gaussianData;
		if (laplacianData != NULL)
			delete laplacianData;
		if (outputBuf != NULL)
			delete outputBuf;
		// Deselect the current rendering context and delete it
		wglMakeCurrent(hDC,NULL);
		wglDeleteContext(hRC);
		
		// Delete the palette if it was created
		if(hPalette != NULL)
			DeleteObject(hPalette);
		
		// Tell the application to terminate after the window
		// is gone.
		PostQuitMessage(0);
		break;
		
		// Window is resized.
	case WM_SIZE:
		// Call our function which modifies the clipping
		// volume and viewport

  		wglMakeCurrent(hDC, hRC);
		GLResize(LOWORD(lParam), HIWORD(lParam));
   		wglMakeCurrent(hDC, NULL);
		break;
		
		
		// The painting function.  This message sent by Windows 
		// whenever the screen needs updating.
	case WM_PAINT:
		{
   			wglMakeCurrent(hDC, hRC);

		 // Call OpenGL drawing code
			GLRenderScene();
			
			SwapBuffers(hDC);
			wglMakeCurrent(hDC, NULL);
			
			// Validate the newly painted client area
			ValidateRect(hwnd, NULL);
			//InvalidateRect(hOutput, NULL, FALSE);
		}
		break;
		
		// Windows is telling the application that it may modify
		// the system palette.  This message in essance asks the 
		// application for a new palette.
	case WM_QUERYNEWPALETTE:
		// If the palette was created.
		if(hPalette)
		{
			int nRet;
			
			// Selects the palette into the current device context
			SelectPalette(hDC, hPalette, FALSE);
			
			// Map entries from the currently selected palette to
			// the system palette.  The return value is the number 
			// of palette entries modified.
			nRet = RealizePalette(hDC);
			
			// Repaint, forces remap of palette in current window
			InvalidateRect(hwnd,NULL,FALSE);
			
			return nRet;
		}
		break;
		
		
		// This window may set the palette, even though it is not the 
		// currently active window.
	case WM_PALETTECHANGED:
		// Don't do anything if the palette does not exist, or if
		// this is the window that changed the palette.
		if((hPalette != NULL) && ((HWND)wParam != hwnd))
		{
			// Select the palette into the device context
			SelectPalette(hDC,hPalette,FALSE);
			
			// Map entries to system palette
			RealizePalette(hDC);
			
			// Remap the current colors to the newly realized palette
			UpdateColors(hDC);
			return 0;
		}
		break;
		
		// Key press, check for arrow keys to do cube rotation.
	case WM_KEYDOWN:
		{
			if ((unsigned char)wParam == 'Z' || (unsigned char)wParam == 'z')
			{
				ZoomIn();
				PostMessage(hwnd, WM_SIZE, SIZE_RESTORED , MAKELPARAM(lastWidth, lastHeight));
			}
			if ((unsigned char)wParam == 'X' || (unsigned char)wParam == 'x')
			{
				ZoomOut();
				PostMessage(hwnd, WM_SIZE, SIZE_RESTORED , MAKELPARAM(lastWidth, lastHeight));
			}
			if(wParam == VK_UP)
				KeyUp();

			if(wParam == VK_DOWN)
				KeyDown();
			
			if(wParam == VK_LEFT)
				KeyLeft();
			
			if(wParam == VK_RIGHT)
				KeyRight();
			isSceneDirty = TRUE;
			updateOutput = TRUE;

			InvalidateRect(hwnd,NULL,FALSE);
		}
		break;

	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
		isButtonDown = TRUE;
		return (DefWindowProc(hwnd, message, wParam, lParam));
		
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
		isButtonDown = FALSE;
		return (DefWindowProc(hwnd, message, wParam, lParam));
		// A menu command
	case WM_COMMAND:
		{
			HWND hwndCtl = (HWND) lParam;
			HMENU hMenu = GetMenu(hwnd);
			UINT iSelection, menuID;

			iSelection = LOWORD(wParam);
			switch(iSelection)
			{
			case ID_FILE_OPEN:
				if (PopFileOpenDlg(hwnd, szFileName, szTitleName))
				{
					if (curModel != NULL)
					{
						wglMakeCurrent(hDC, hRC);
						glDeleteLists(curModel->GetListID(), 1);
 						wglMakeCurrent(hDC, NULL);
						delete curModel;
					}
					curModel = new Model();
					if (!PopFileRead(hOutput, szFileName, curModel))
					{
						OkMessage(hwnd, "Could not read file %s!", szTitleName);
						szFileName[0] = '\0';
						szTitleName[0] = '\0';
					}
					wglMakeCurrent(hDC, hRC);
					GLSetupRC();
 					wglMakeCurrent(hDC, NULL);
					isSceneDirty = TRUE;
					updateOutput = TRUE;
				}
				DoCaption(hwnd, szTitleName);
				return 0;
			case ID_OUTPUT_SAVE:
				if (PopFileSaveDlg(hwnd, szFileName, szTitleName))
				{
					DoCaption(hOutput, szTitleName);
					if (OutputDepthFile(szFileName))
					{
						return 1;
					}
					else 
						OkMessage(hwnd, "Could not write file %s", szTitleName);
				}
				return 0;
			case ID_OUTPUT_REFRESH:
					InvalidateRect(hwnd,NULL,FALSE);
					refreshOutput = TRUE;
					//isSceneDirty = TRUE;
					//updateOutput = TRUE;
				return 0;
			case ID_VIEW_ZBUFFER:
				if (viewMode == 0)
				{
					CheckMenuItem(hMenu, iSelection, MF_UNCHECKED);
					viewMode = 1;
				}
				else
				{
					CheckMenuItem(hMenu, iSelection, MF_CHECKED);
					viewMode = 0;
				}
				if (hOutput == NULL)
					CreateOutputWindow(hwnd,hInstance);
				updateOutput = TRUE;
				break;
			case ID_VIEW_PMC:
			case ID_VIEW_GAUSSIAN:
			case ID_VIEW_LAPLACIAN:
				menuID = iSelection - ID_VIEW_PMC;
				if (menuChecked[menuID] == TRUE)
				{
					CheckMenuItem(hMenu, iSelection, MF_UNCHECKED);
					//viewMode = 0;
					menuChecked[menuID] = FALSE;
				}
				else
				{
					CheckMenuItem(hMenu, iSelection, MF_CHECKED);
					//viewMode = 1;
					menuChecked[menuID] = TRUE;
				}
				if (hOutput == NULL)
					CreateOutputWindow(hwnd,hInstance);
				updateOutput = TRUE;
				break;
			case ID_OPTIONS_THRESHOLDS:
				hDlgModeless = CreateDialogParam( hInstance,
					MAKEINTRESOURCE(IDD_DIALOG_THRESHOLDS),
					hwnd,
					(DLGPROC) ThresholdProc,
					(LPARAM)pThreshold);
				break;
				// Exit the program
			case ID_FILE_EXIT:
				DestroyWindow(hwnd);
				break;
				
				// Display the about box
			case ID_HELP_ABOUT:
				DialogBox (hInstance,
					MAKEINTRESOURCE(IDD_DIALOG_ABOUT),
					hwnd,
					(DLGPROC)AboutDlgProc);
				break;
			}
		}
		break;
		
		
	default:   // Passes it on if unproccessed
		return (DefWindowProc(hwnd, message, wParam, lParam));
		
	}
	
    return (0L);
}


BOOL OutputDepthFile(char *fileName)
{
	BOOL	rets;
	int error, size;
	PgmFileInfoPtr destInfo = NULL;
	
	rets = FALSE;
	destInfo = (PgmFileInfoPtr) new PgmFileInfo;
	if (destInfo == NULL)
		goto CLEANUP;
	destInfo->width = lastWidth;
	destInfo->height = lastHeight;
	destInfo->maxGrayValue = 255;
	size = lastWidth * lastHeight;
	destInfo->data = (unsigned char*) outputBuf;
	if (destInfo->data == NULL)
		goto CLEANUP;
	error = PgmWriteFile(fileName, destInfo);
	if (error == 0)
		rets = TRUE;
CLEANUP:
	if (destInfo)
		delete destInfo;
	return rets;
}